// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_CongestionController_hpp
#define anna_comm_CongestionController_hpp

#include <utility>

#include <anna/config/defines.hpp>
#include <anna/core/Singleton.hpp>
#include <anna/core/mt/NRMutex.hpp>
#include <anna/core/util/Average.hpp>

namespace anna {

namespace xml {
class Node;
}

namespace comm {

class ClientSocket;

namespace handler {
class LocalConnection;
}

/**
   Gestor de congestion autonomo.

Su funcionamiento a grandes rasgos es el siguiente:

   -# El campo \em getsockopt (...SO_RCVBUF...) nos da el tamao maximo del buffer de recepcion
   para un socket dado. Segun hemos visto puede varias desde 32K hasta 80 K, dependiendo del sistema
   operativo y la configuracion. Independientemente del valor particular este sera nuestro limite
   maximo (100%), ya que si la longitud de la cola de espera alcanza este valor, el socket dejara de
   estar disponible para todos los clientes.
   -# El metodo \em ioctl (....FIONREAD...) nos da el numero de bytes pendientes de tratar en el socket.
   -# La carga del socket estara definida por (2) * 100 / (1).
   -# Con esta clase podemos limitar la carga maxima soportada por los sockets de nuestra
   aplicacion. Por defecto sera 60% del valor (1), un valor mas conservador podria ser 40%.
   -# La zona sin congestion estara entre 0% y el valor (4). La zona de congestion estara entre (valor (4) + 1)%  y el 100%.
   -# La zona de congestion se divide, a su vez, en 4 zonas logicas: Si la carga del socket esta en el
   primer 25% se descartan mensajes con una probabilidad del 25%. Si la carga esta en el segundo 25% se
   descartan mensajes con una probabilidad del 50%. Si la carga esta entre el 51% y el 85% se descartan
   paquetes con una probabilidad del 85% y si esta entre el 86% y el 100% se descartan los paquetes con
   una probabilidad del 99%.

Ventajas:

   \li Trata cada socket de forma independiente, con lo que ajustamos mucho mejor el rendimiento de
   nuestra aplicacion.
   \li No hay que imaginar los 4 niveles de a8_controlniveles. Solo tendremos que establecer el nivel
   de carga maximo y el resto lo hace el proceso por si mismo. Sea como sea, es mucho mas facil ajustar
   un unico valor que los 8 o 10 necesarios con el sistema anterior.
   \li No requiere ninguna configuracion en base de datos.
*/
class CongestionController : public Singleton <CongestionController> {
public:
  /**
   * Nivel mximo de congestin.
   */
  static const int MaxLevel = 4;

  /**
     Posibles resultados de CongestionController::getAdvice.
  */
  struct Advice {
    enum _v {
      None, /**< No hay datos para tomar una decision.  */
      Process, /**< Se puede procesar el mensaje sin contribuir a la congestion */
      Discard /**< Se aconseja descartar para limitar la congestion */
    };
  };

  /**
     Posibles formas de calcular la carga que hara aconsejar el descarte o aceptacion de un mensaje.
  */
  struct Mode {
    enum _v {
      Auto, /**< Escoge la forma de calculo que mejor se adapte al modo de compilacion. Normalmente
                El calculo de tipo Local sera mas util en modo MT, mientras que el Global puede ser
                mas aconsejable en modo ST.
                */
      Local, /**< El calculo de la carga se calcula de forma independiente para cada Socket */
      Global /**< El calculo de la carga se calcula de forma conjunta para todos los Sockets */
    };
  };

  typedef std::pair <int, int> Workload;

  static const int DefaultLimit = 60;

  /**
   * Valor mximo que se puede asignar al control de bytes pendientes de tratar.
   */
  static const int MaxPendingBytes;

  /**
     Devuelve el modo que estamos usando para calcular la carga.
     \return El modo que estamos usando para calcular la carga.
  */
  Mode::_v getMode() const throw() { return a_mode; }

  /**
   * Devuelve el nmero total de mensajes recibididos.
   * \return el nmero total de mensajes recibididos.
   */
  int getMessageCounter() const throw() { return a_messageCounter; }

  /**
   * Devuelve el nmero de mensajes tratados correctamente.
   * \return el nmero de mensajes tratados correctamente.
   */
  int getSuccessCounter() const throw() { return a_messageCounter - a_discardedCounter; }

  /**
     Establece el limite de carga que vamos a imponer.
     \param limit Limite de carga. Debe ser un valor entre 0 y 100.
  */
  void setLimit(const int limit) throw();

  /**
     Establece el modo en que vamos a calcular la carga.
     \param mode Modo usado para calcular la carga.
  */
  void setMode(const Mode::_v mode) throw() {
    if((a_mode = mode) == Mode::Auto) {
      WHEN_MULTITHREAD(a_effectiveMode = Mode::Local);
      WHEN_SINGLETHREAD(a_effectiveMode = Mode::Global);
    } else
      a_effectiveMode = mode;
  }

  /**
     Devuelve \em true si esta instancia ha recibido datos o \em false en otro caso.
     \return \em true si esta instancia ha recibido datos o \em false en otro caso.
  */
  bool isEmpty() const throw() { return a_avgWorkload.isEmpty(); }

  /**
   *  Establece el n mximo de bytes que deberan tener los socket en la cola de entrada.
   * \param maxPendingBytes Nmero mximo de bytes que debera tener los socket en la cola de entrada.
   * \warning El valor indicado debe estar dentro del siguiente mbito:
   * \code
   * [#comm::Communicator::getReadingChunkSize, min (#comm::Communicator::getReadingChunkSize * 4, #MaxPendingBytes)]
   * \endcode
   */
  void setMaxPendingBytes(const int maxPendingBytes) throw(RuntimeException);

  /**
     Devuelve el consejo sobre si debemos tratar/o no el ultimo mensaje recibido
     por el ClientSocket recibido como parametro.
     \param clientSocket Socket cliente por el que hemos recibido el ultimo mensaje.
     \return Advice::Process para indicar que debemos procesar el mensaje
     , Advice::None para indicar que todavia no tiene datos para tomar una
     decision clara o Advice::Discard para indicar que debemos descartar.
  */
  Advice::_v getAdvice(const ClientSocket& clientSocket) throw();

  /**
     Devuelve el consejo sobre si debemos tratar/o no el ultimo mensaje recibido
     por el ClientSocket recibido como parametro.
     \param clientSocket Socket cliente por el que hemos recibido el ultimo mensaje.
     \return Advice::Process para indicar que debemos procesar el mensaje
     , Advice::None para indicar que todavia no tiene datos para tomar una
     decision clara o Advice::Discard para indicar que debemos descartar.
  */
  Advice::_v getAdvice(const ClientSocket* clientSocket) throw() {
    return (clientSocket == NULL) ? Advice::Process : getAdvice(*clientSocket);
  }

  /**
   * Devuelve informacin sobre las estadsticas de carga en las que se bas para hacer
   * los clculos.
   *
   * \warning Slo debera invocarse a este mtodo despus de invocar a #getAdvice.
   * \warning Este mtodo no es MT-estricto, por lo que en un entorno MT los valores
   * sobre los que se calcul el ltimo #getAdvice y los datos obtenidos con este mtodo
   * pueden haber variado ligeramente.
   *
   * \see #getLoad
   * \see #getLevel
   */
  Workload getAccumulatedWorkload() const throw();

  /**
   * Devuelve informacin sobre las estadsticas de carga del socket recibido como parmetro.
   *
   *
   * \param clientSocket Socket del que se quiere obtener el estado actual de ocupacin.
   *
   * \warning Slo debera invocarse a este mtodo despus de invocar a #getAdvice.
   * \warning Este mtodo no es MT-estricto, por lo que en un entorno MT los valores
   * sobre los que se calcul el ltimo #getAdvice y los datos obtenidos con este mtodo
   * pueden haber variado ligeramente.
   *
   * \see #getLoad
   * \see #getLevel
   */
  Workload getCurrentWorkload(const ClientSocket& clientSocket) const throw();

  /**
     Devuelve un documento XML con la informacion relevante sobre esta clase.
     \return un documento XML con la informacion relevante sobre esta clase.
  */
  xml::Node* asXML(xml::Node* parent) const throw();

  /**
   * Extrae la carga media que soporta es proceso. La carga es un nmero en [0, 100].
   * \return la carga media que soporta es proceso. La carga es un nmero en [0, 100].
   */
  static int getLoad(const Workload& workload) throw() { return workload.second; }

  /**
   * Extrae el nivel de carga en la que est el proceso. El nivel de carga es un nmero en [0, 4].
   * \return el nivel de carga en la que est el proceso. El nivel de carga es un nmero en [0, 4].
   */
  static int getLevel(const Workload& workload) throw() { return workload.first; }

private:
  static const Millisecond DelayTrace;
  static const int UnusedPendingBytes = -1;
  static const Millisecond HeartBeat;

  //----------------------------------------------------------------------------------
  // a_limit: % de ocupacion maxima al que vamos a limitar los canales.
  // a_discardLevel: Valor maximo del % de cada nivel de descarte. 0 = 25, 1 = 50,
  // 2 = 85, 3 = 99
  //----------------------------------------------------------------------------------
  int a_limit;
  int a_discardLevel [MaxLevel];
  int a_percentage [MaxLevel];
  NRMutex a_mutex;
  Mode::_v a_mode;
  Mode::_v a_effectiveMode;
  Average <unsigned int> a_avgWorkload;
  unsigned int a_messageCounter;
  unsigned int a_discardedCounter;
  mutable Millisecond a_timeTrace;
  int a_maxPendingBytes;
  int a_incomingSocketCounter;
  Millisecond a_tickTime;

  void incrementIncomingSocket() throw(RuntimeException);
  void decrementIncomingSocket() throw(RuntimeException);

  CongestionController();
  CongestionController(const CongestionController&);

  int calculeWorkload(const ClientSocket&) const throw();

  friend class Singleton <CongestionController>;

  friend class ClientSocket;
  // getUpdatePeriod

  friend class handler::LocalConnection;
  // incrementIncomingSocket, decrementIncomingSocket
};

}
}

#endif

