// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_oam_CounterScope_hpp
#define anna_core_oam_CounterScope_hpp

#include <anna/core/functions.hpp>
#include <anna/core/RuntimeException.hpp>
#include <anna/core/mt/Mutex.hpp>

#include <anna/core/oam/Counter.hpp>

namespace anna {

class Guard;

namespace xml {
class Node;
}

namespace oam {


/**
   mbito de contadores. Contiene un grupo logico de contadores.
*/
class CounterScope  : public Mutex {
public:

  static const int MaxCounter = 1000; /**< Numero maximo de contador por cada ambito */

  /**
     Destructor.
  */
  ~CounterScope();

  /**
     Devuelve el nombre logico del contador, que coincidira con el indicado a la hora
     de crear el contador mediante el metodo CounterScope::create.
     \return El nombre logico del contador.
  */
  const std::string& getName() const throw() { return a_name; }

  /**
     Devuelve el identificador del ambito
     \return El identificador del ambito.
  */
  const int getId() const throw() { return a_id; }

  /**
     Crea un nuevo contador.
     \param counter Numero logico del contador a crear. Debera ser menor de MaxCounter.
     \param name Nombre logico del ambito.
  */
  void create(const int counter, const char* name) throw(RuntimeException);

  /**
     Devuelve una cadena con la informacion relevante de este objeto.
     \return Una cadena con la informacion relevante de este objeto.
  */
  std::string asString() const throw();

  /**
   * Devuelve la informacin relevante de esta instancia en un documento XML.
   * \return la informacin relevante de esta instancia en un documento XML.
   */
  xml::Node* asXML(xml::Node* parent) const throw(RuntimeException);

protected:
  /**
   * Incrementa el contador recibido como parmetro.
   * \warning La invocacin a este mtodo deber hacerse sobre una seccin crtica que proteja
   * a este mbito.
   * \return El valor actual del contador.
   */
  Counter::type_t increment(const int counter, const Counter::type_t value) throw(RuntimeException);

  /**
   * Establece el valor del contador recibido como parmetro.
   * \warning La invocacin a este mtodo deber hacerse sobre una seccin crtica que proteja
   * a este mbito.
   * \return El valor actual del contador.
   */
  Counter::type_t assign(const int counter, const Counter::type_t value) throw(RuntimeException);

  /**
   * Devuelve el valor actual del contador pasado como parmetro.
   * \param counter Identificador del contedor cuyo valor queremos obtener.
   */
  Counter::type_t getValue(const int counter) const throw(RuntimeException);

  /**
   * Devuelve la instancia del contador. Puede ser NULL.
   * \return la instancia del contador.
   */
  const Counter* getCounter(const int counter) const throw(RuntimeException);

public:

  /**
   * Devuelve el valor actual acumulado del contador pasado como parmetro.
   * \param counter Identificador del contedor cuyo valor acumulado queremos obtener.
   */
  U64 getAccValue(const int counter) const throw(RuntimeException);

  /**
   * Resetea los valores acumulados totales de los contadores incluidos en el mbito.
   * \return Numero de contadores afectados que tenian un acumulado no nulo.
   */
  int resetAccValues() throw(RuntimeException);

private:
  const int a_id;
  const std::string a_name;
  Counter* a_counters [MaxCounter];

  CounterScope(const int id, const char* name) :
    Mutex(Mutex::Mode::Recursive),
    a_id(id),
    a_name(name) {
    anna_memset(a_counters, 0, sizeof(a_counters));
  }

  friend class Safe;
  friend class Handler;
  friend class Module;
};

}
}

#endif

