// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <unistd.h>

#include <iostream>

#include <anna/core/functions.hpp>

#include <anna/core/mt/ThreadManager.hpp>
#include <anna/core/mt/Semaphore.hpp>
#include <anna/core/mt/Thread.hpp>
#include <anna/core/tracing/TraceMethod.hpp>

using namespace std;
using namespace anna;

anna_assign_enum(ThreadManager::Mode) = { "None", "Unlimit", "ExceptionWhenFull", "LockWhenFull", NULL };

ThreadManager::ThreadManager(const char* name, const Mode::_v mode, const int maxSize, const int flags) :
  a_name(name),
  a_mode(mode),
  a_maxSize(maxSize),
  a_semaphore(NULL),
  a_threadFlags(flags) {
  if(a_mode == Mode::LockWhenFull)
    a_semaphore = new Semaphore(0);
}

ThreadManager::ThreadManager(const char* name, const int flags) :
  a_name(name),
  a_mode(Mode::Unlimit),
  a_maxSize(-1),
  a_semaphore(NULL),
  a_threadFlags(flags) {
  if(a_mode == Mode::LockWhenFull)
    a_semaphore = new Semaphore(0);
}

ThreadManager::~ThreadManager() {
#ifdef _MT

  try {
    // Para que no intenten sacar el nombre del ThreadManager que los cre
    for(thread_iterator ii = thread_begin(), maxii = thread_end(); ii != maxii; ii ++) {
      thread(ii)->a_manager = NULL;
    }

    if(a_semaphore) {
      a_semaphore->signal();
      delete a_semaphore;
    }
  } catch(RuntimeException& ex) {
    ex.trace();
  }

#else

  if(a_semaphore)
    delete a_semaphore;

#endif
}

Thread* ThreadManager::createThread()
throw(RuntimeException) {
  Thread* result;

  if(a_mode == Mode::None) {
    string msg(asString());
    msg += " | Invalid mode";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  if(a_mode != Mode::Unlimit && a_maxSize <= 0) {
    string msg(asString());
    msg += " | Invalid max thread number";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  Guard guard(this, "anna::ThreadManager::createThread");
  const int size = Recycler <Thread>::size();

  switch(a_mode) {
  case Mode::ExceptionWhenFull:

    if(a_maxSize == size) {
      string msg(asString());
      msg += " | No available threads";
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    // No hay "break" para que siga procesando
  case Mode::Unlimit:
    result = Recycler <Thread>::create();
    break;
  case Mode::LockWhenFull:

    if(a_maxSize == size) {
      guard.deactivate();
      LOGDEBUG(
        string msg(asString());
        msg += " | Waiting for thread release";
        Logger::debug(msg, ANNA_FILE_LOCATION);
      );
      a_semaphore->wait();
      LOGDEBUG(
        string msg(asString());
        msg += " | Achieve thread release";
        Logger::debug(msg, ANNA_FILE_LOCATION);
      );
      Guard reopen(this, "anna::ThreadManager::createThread (after signal)");
      result = Recycler <Thread>::create();
    } else
      result = Recycler <Thread>::create();

    break;
  default: break;
  }

  result->a_manager = this;
  result->setFlags(a_threadFlags);
  LOGDEBUG(
    string msg("ThreadManager::createThread | ");
    msg += asString();
    msg += " | ";
    msg += result->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  return result;
}

void ThreadManager::join()
throw(RuntimeException) {
  LOGDEBUG(
    string msg("ThreadManager::join (init) | ");
    msg += asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
#ifdef _MT
  lock();
  const pthread_t self(pthread_self());
  pthread_t* threads = new pthread_t [Recycler <Thread>::size()];
  int index = 0;

  for(thread_iterator ii = thread_begin(), maxii = thread_end(); ii != maxii; ii ++) {
    threads [index] = thread(ii)->getId();

    if(pthread_equal(threads [index], self) != 0) {
      string msg(asString());
      msg += " | ";
      thread(ii)->asString();
      msg += " | Threads owns to this Manager";
      unlock();
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    index ++;
  }

  unlock();
  int errorCode;

  for(int ii = 0; ii < index; ii ++) {
    if((errorCode = pthread_join(threads [ii], NULL)) != 0) {
      string msg(asString());
      msg += " | Bad join";
      throw RuntimeException(msg, errorCode, ANNA_FILE_LOCATION);
    }
  }

  delete [] threads;
#endif
  LOGDEBUG(
    string msg("ThreadManager::join (final) | ");
    msg += asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

void ThreadManager::releaseThread(Thread* thread)
throw(RuntimeException) {
  if(thread->isRunning() == true) {
    string msg(thread->asString());
    msg += " | Still activated";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  if(a_mode == Mode::LockWhenFull) {
    Guard guard(this, "anna::ThreadManager::releaseThread");
    // No hace falta acceder mediate SafeRecycler porque ya tenemos una SSCC establecida.
    const int size = Recycler <Thread>::getSize();
    Recycler <Thread>::release(thread);

    if(size == a_maxSize)
      a_semaphore->signal();
  } else
    SafeRecycler <Thread>::release(thread);

  LOGDEBUG(
    string msg("ThreadManager::releaseThread | ");
    msg += asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  )
}

string ThreadManager::asString() const
throw() {
  string result("anna::ThreadManager { Name: ");
  result += a_name;
  result += " | Mode: ";
  result += Mode::asCString(a_mode);
  result += " | MaxSize: ";

  if(a_maxSize > 0)
    result += functions::asString(a_maxSize);
  else
    result += "<unlimited>";

  result += functions::asText(" | Size: ", Recycler <Thread>::size());
  return result += " }";
}

