// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/diameter.comm/ClientSession.hpp>
#include <anna/diameter.comm/Entity.hpp>
#include <anna/diameter.comm/Engine.hpp>
#include <anna/diameter.comm/OamModule.hpp>
#include <anna/diameter/codec/functions.hpp>
#include <anna/diameter/helpers/base/defines.hpp>
#include <anna/core/functions.hpp>
#include <anna/statistics/Engine.hpp>
#include <anna/time/functions.hpp>

#include <anna/core/tracing/Logger.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/tracing/TraceMethod.hpp>

// STL
#include <vector>

using namespace anna;
using namespace anna::diameter;
using namespace anna::diameter::comm;


void Server::initialize() throw() {
  a_parent = NULL;
  a_engine = NULL;
  a_clientSessions.clear(); // importante (el recycler creo que no lo tocaba)
  a_available = false;
  a_maxClientSessions = 1; // mono client connection
  a_lastIncomingActivityTime = (anna::Millisecond)0;
  a_lastOutgoingActivityTime = (anna::Millisecond)0;
  a_lastUsedResource = NULL;
}

void Server::initializeStatisticResources() throw() {
  std::string accName = "remote server '";
  accName += anna::functions::socketLiteralAsString(a_socket.first, a_socket.second);
  accName += "' from origin-realm '";
  accName += a_engine ? a_engine->getOriginRealmName() : "unknown"; // it should be known (createServer)
  accName += "' and origin-host '";
  accName += a_engine ? a_engine->getOriginHostName() : "unknown"; // it should be known (createServer)
  accName += "'";
  a_messageStatistics.initialize(accName);
}

void Server::resetStatistics() throw() {
  a_messageStatistics.getAccumulator()->reset();
}

void Server::updateProcessingTimeStatisticConcept(const double &value, const anna::diameter::CommandId &cid) throw() {
  a_messageStatistics.process(MessageStatistics::ConceptType::SentRequestProcessingTime, cid, value);
  LOGDEBUG(anna::Logger::debug(a_messageStatistics.getAccumulator()->asString(), ANNA_FILE_LOCATION));
}

void Server::updateReceivedMessageSizeStatisticConcept(const double &value, const anna::diameter::CommandId &cid) throw() {
  a_messageStatistics.process(MessageStatistics::ConceptType::ReceivedMessageSize, cid, value);
  LOGDEBUG(anna::Logger::debug(a_messageStatistics.getAccumulator()->asString(), ANNA_FILE_LOCATION));
}


void Server::assertReady() throw(anna::RuntimeException) {
  if(a_clientSessions.size() != a_maxClientSessions) {
    std::string msg(asString());
    msg += " | Non-configured server: you must add the remaining client-sessions before any operation (bind, send, etc.)";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }
}


void Server::addClientSession(int socketId)
throw(anna::RuntimeException) {
  if(a_clientSessions.size() == a_maxClientSessions) {
    LOGDEBUG
    (
      std::string msg = "Maximum number of client-sessions reached for this server (";
      msg += anna::functions::asString(a_maxClientSessions);
      msg += ").";
      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
    );
    return;
  }

  if(!a_engine)
    throw anna::RuntimeException("Invalid engine reference (NULL)", ANNA_FILE_LOCATION);

  ClientSession *s = a_engine->createClientSession(this, socketId);
  a_clientSessions.push_back(s);
}

int Server::getOTARequests() const throw() {
  int result = 0;

  for(std::vector<ClientSession*>::const_iterator it = begin(); it != end(); it++)
    result += (*it)->getOTARequests();

  return result;
}


bool Server::send(const Message* message, int socketId) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Server", "send", ANNA_FILE_LOCATION));
  assertReady();
  bool fixedSocket = (socketId != -1);
  int clientSessions = getNumberOfClientSessions();

  for(int k = 0; k < clientSessions; k++) {    // try round-robin only over one cycle,
    // no matter where you are: don't repeat same socket
    if(fixedSocket)
      a_lastUsedResource = a_engine->findClientSession(a_socket.first /*ip*/, a_socket.second /*port*/, socketId); // exception if not found
    else {
      if(a_deliveryIterator == end()) a_deliveryIterator = begin();

      a_lastUsedResource = (*a_deliveryIterator);

      if(clientSessions > 1) a_deliveryIterator++;  // Round robin
    }

    try {
      // Send:
      a_lastUsedResource->send(message);
      return true; // no matter if response is NULL (answers, i.e.) or not.
    } catch(anna::RuntimeException &ex) {
      LOGDEBUG(
        std::string msg = ex.getText();
        msg += " | Send failed on socket id ";
        msg += anna::functions::asString(a_lastUsedResource->getSocketId());
        anna::Logger::debug(msg, ANNA_FILE_LOCATION);
      );
    }

    // Specific socket sending doesn't try
    if(fixedSocket) return false;
  }

  return false;
}


bool Server::broadcast(const Message* message) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Server", "broadcast", ANNA_FILE_LOCATION));
  assertReady();
  bool allok = true;

  for(std::vector<ClientSession*>::iterator it = begin(); it != end(); it++) {
    try {
      (*it)->send(message);
    } catch(anna::RuntimeException &ex) {
      ex.trace();
      allok = false;
    }
  }

  return allok;
}


bool Server::bind() throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Server", "bind", ANNA_FILE_LOCATION));
  assertReady();
  a_deliveryIterator = begin();
  bool result = true; // all OK return

  for(std::vector<ClientSession*>::iterator it = begin(); it != end(); it++) {
    try {
      (*it)->bind();
    } catch(anna::RuntimeException &ex) {
      ex.trace();
      result = false;
    }
  }

  return result;
}

void Server::raiseAutoRecovery(bool autoRecovery) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Server", "raiseAutoRecovery", ANNA_FILE_LOCATION));
  assertReady();
  a_deliveryIterator = begin();

  for(std::vector<ClientSession*>::iterator it = begin(); it != end(); it++)
    (*it)->setAutoRecovery(autoRecovery);
}

void Server::setClassCodeTimeout(const ClassCode::_v v, const anna::Millisecond & millisecond) throw() {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Server", "setClassCodeTimeout", ANNA_FILE_LOCATION));
  assertReady();

  for(std::vector<ClientSession*>::iterator it = begin(); it != end(); it++) {
    try {
      (*it)->setClassCodeTimeout(v, millisecond);
    } catch(anna::RuntimeException &ex) {
      ex.trace();
    }
  }
}


// Private close/destroy method
void Server::close(bool destroy) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Server", "close", ANNA_FILE_LOCATION));

  if(!a_engine)
    throw anna::RuntimeException("Invalid engine reference (NULL)", ANNA_FILE_LOCATION);

  assertReady();

  for(std::vector<ClientSession*>::iterator it = begin(); it != end(); it++)
    a_engine->closeClientSession(*it, destroy);
}


void Server::childIdle() const throw() {
  // Check father entity idleness:
  if(idle()) a_parent->childIdle();
}


void Server::hide() throw() {
  for(std::vector<ClientSession*>::iterator it = begin(); it != end(); it++)
    (*it)->hide();
}

void Server::show() throw() {
  for(std::vector<ClientSession*>::iterator it = begin(); it != end(); it++)
    (*it)->show();
}

bool Server::hidden() const throw() {
  for(std::vector<ClientSession*>::const_iterator it = begin(); it != end(); it++)
    if((*it)->shown()) return false;

  return true;
}
bool Server::shown() const throw() {
  for(std::vector<ClientSession*>::const_iterator it = begin(); it != end(); it++)
    if((*it)->hidden()) return false;

  return true;
}



std::string Server::socketAsString() const throw() {
  std::string result = getAddress();
  result += ":";
  result += anna::functions::asString(getPort());
  return result;
}


std::string Server::asString() const throw() {
  std::string result("diameter::comm::Server { ");
  result += " | Parent Entity: ";
  result += a_parent->getDescription();
  result += " | Server Address: ";
  result += a_socket.first;
  result += " | Server Port: ";
  result += anna::functions::asString(a_socket.second);
  result += " | Available: ";
  result += a_available ? "yes" : "no";
  result += " | Max client-sessions supported: ";
  result += anna::functions::asString(a_maxClientSessions);
  result += " | Currently configured client-sessions: ";
  result += anna::functions::asString(a_clientSessions.size());
  result += anna::functions::asString(" | OTA requests: %d%s", getOTARequests(), idle() ? " (idle)" : "");
  result += " | Last Incoming Activity Time: ";
  result += a_lastIncomingActivityTime.asString();
  result += " | Last Outgoing Activity Time: ";
  result += a_lastOutgoingActivityTime.asString();
  result += " | Hidden: ";
  result += (hidden() ? "yes" : "no");
  result += "\n";
  result += a_messageStatistics.getAccumulator()->asString();

  for(std::vector<ClientSession*>::const_iterator it = begin(); it != end(); it++) {
    result += "\n";
    result += (*it)->asString();
  }

  return result;
}

anna::xml::Node* Server::asXML(anna::xml::Node* parent) const throw() {
  anna::xml::Node* result = parent->createChild("diameter.Server");
  result->createAttribute("ParentEntity", a_parent->getDescription());
  result->createAttribute("ServerAddress", a_socket.first);
  result->createAttribute("ServerPort", anna::functions::asString(a_socket.second));
  result->createAttribute("Available", a_available ? "yes" : "no");
  result->createAttribute("MaxClientSessionsSupported", anna::functions::asString(a_maxClientSessions));
  result->createAttribute("CurrentlyConfiguredClientSessions", anna::functions::asString(a_clientSessions.size()));
  result->createAttribute("OTArequests", anna::functions::asString("%d%s", getOTARequests(), idle() ? " (idle)" : ""));
  result->createAttribute("LastIncomingActivityTime", a_lastIncomingActivityTime.asString());
  result->createAttribute("LastOutgoingActivityTime", a_lastOutgoingActivityTime.asString());
  result->createAttribute("Hidden", hidden() ? "yes" : "no");
  // Statistics
  anna::xml::Node* stats = result->createChild("Statistics");
  a_messageStatistics.getAccumulator()->asXML(stats);
  anna::xml::Node* clientSessions = result->createChild("Server.ClientSessions");

  for(std::vector<ClientSession*>::const_iterator it = begin(); it != end(); it++)
    (*it)->asXML(clientSessions);

  return result;
}


void Server::eventPeerShutdown(const ClientSession *clientSession) throw() {
  // Inform father entity:
  a_parent->eventPeerShutdown(clientSession);
}

void Server::eventRequestRetransmission(const ClientSession* clientSession, Message *request) throw() {
  // Inform father entity:
  a_parent->eventRequestRetransmission(clientSession, request);
}

void Server::eventResponse(const Response& response) throw(anna::RuntimeException) {
  // Inform father entity:
  a_parent->eventResponse(response);
}

void Server::eventRequest(ClientSession *clientSession, const anna::DataBlock & request) throw(anna::RuntimeException) {
  // Inform father entity:
  a_parent->eventRequest(clientSession, request);
}

void Server::eventUnknownResponse(ClientSession *clientSession, const anna::DataBlock & response) throw(anna::RuntimeException) {
  // Inform father entity:
  a_parent->eventUnknownResponse(clientSession, response);
}

void Server::eventDPA(ClientSession *clientSession, const anna::DataBlock & response) throw(anna::RuntimeException) {
  // Inform father entity:
  a_parent->eventDPA(clientSession, response);
}


void Server::availabilityLost() throw() {
  a_available = false;
  std::string socket = anna::functions::socketLiteralAsString(a_socket.first, a_socket.second);
  LOGDEBUG(
    std::string msg = "diameter::comm::Server { Socket: ";
    msg += socket;
    msg += " } has lost its availability";
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  oamModule.activateAlarm(OamModule::Alarm::c_LostAvailabilityOverServerDefinedAs__s__, socket.c_str());
  oamModule.count(OamModule::Counter::LostAvailabilityOverServer);
  a_engine->availabilityLost(this);
  a_parent->refreshAvailability();
}


void Server::availabilityRecovered() throw() {
  a_available = true;
  std::string socket = anna::functions::socketLiteralAsString(a_socket.first, a_socket.second);
  LOGDEBUG(
    std::string msg = "diameter::comm::Server { Socket: ";
    msg += socket;
    msg += " } has recovered its availability";
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  oamModule.cancelAlarm(OamModule::Alarm::c_LostAvailabilityOverServerDefinedAs__s__, socket.c_str());
  oamModule.count(OamModule::Counter::RecoveredAvailabilityOverServer);
  a_engine->availabilityRecovered(this);
  a_parent->refreshAvailability();
}



bool Server::refreshAvailability() throw() {
  // Here available
  if(a_available) {  // check not-bound state for all client-sessions:
    bool isolate = true;

    for(std::vector<ClientSession*>::const_iterator it = begin(); it != end(); it++)

      //if ((*it)->getState() == ClientSession::State::Bound) { isolate = false; break; }
      if((*it)->getState() != ClientSession::State::Closed) { isolate = false; break; }

    if(isolate) {
      availabilityLost();
      return true;
    }

    return false;
  }

  // Here not available
  for(std::vector<ClientSession*>::const_iterator it = begin(); it != end(); it++)
    if((*it)->getState() == ClientSession::State::Bound) {
      availabilityRecovered();
      return true;
    }

  return false;
}


//------------------------------------------------------------------------------
//---------------------------------------- Server::updateIncomingActivityTime()
//------------------------------------------------------------------------------
void Server::updateIncomingActivityTime() throw() {
  a_lastIncomingActivityTime = anna::functions::millisecond();
  LOGDEBUG
  (
    std::string msg = "Updated INCOMING activity on server (milliseconds unix): ";
    msg += anna::functions::asString(a_lastIncomingActivityTime.getValue());
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  a_parent->updateIncomingActivityTime();
}


//------------------------------------------------------------------------------
//---------------------------------------- Server::updateOutgoingActivityTime()
//------------------------------------------------------------------------------
void Server::updateOutgoingActivityTime(void) throw() {
  a_lastOutgoingActivityTime = anna::functions::millisecond();
  LOGDEBUG
  (
    std::string msg = "Updated OUTGOING activity on server (milliseconds unix): "; msg += anna::functions::asString(a_lastOutgoingActivityTime.getValue());
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  a_parent->updateOutgoingActivityTime();
}
